package envvar

import (
	"os"

	"github.com/gofiber/fiber/v3"
)

const hAllow = fiber.MethodGet + ", " + fiber.MethodHead

// Config defines the config for middleware.
type Config struct {
	// ExportVars specifies the environment variables that should export
	ExportVars map[string]string
}

type EnvVar struct {
	Vars map[string]string `json:"vars"`
}

func (envVar *EnvVar) set(key, val string) {
	envVar.Vars[key] = val
}

func New(config ...Config) fiber.Handler {
	var cfg Config
	if len(config) > 0 {
		cfg = config[0]
	}

	return func(c fiber.Ctx) error {
		method := c.Method()
		if method != fiber.MethodGet && method != fiber.MethodHead {
			c.Set(fiber.HeaderAllow, hAllow)
			return fiber.ErrMethodNotAllowed
		}

		envVar := newEnvVar(cfg)
		varsByte, err := c.App().Config().JSONEncoder(envVar)
		if err != nil {
			return c.Status(fiber.StatusInternalServerError).SendString(err.Error())
		}
		c.Set(fiber.HeaderContentType, fiber.MIMEApplicationJSONCharsetUTF8)
		return c.Send(varsByte)
	}
}

func newEnvVar(cfg Config) *EnvVar {
	vars := &EnvVar{Vars: make(map[string]string)}

	if len(cfg.ExportVars) == 0 {
		// do not expose environment variables when no configuration
		// is supplied to prevent accidental information disclosure
		return vars
	}

	for key, defaultVal := range cfg.ExportVars {
		vars.set(key, defaultVal)
		if envVal, exists := os.LookupEnv(key); exists {
			vars.set(key, envVal)
		}
	}

	return vars
}
